import logging
from logging import (CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET)


def get_logger(level='ERROR', module_name='bioat', func_name='', log_mode='s', log_path=None):
    level = level.upper()

    dt_level = {
        k: v
        for k, v in zip(
            ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
            [CRITICAL, ERROR, WARNING, INFO, DEBUG, NOTSET]
        )
    }

    if func_name:
        logger = logging.getLogger(f'{module_name}.{func_name}')
    else:
        logger = logging.getLogger(f'{module_name}')

    logger.setLevel(level=dt_level[level])
    if log_mode == 's':
        logger_handler = logging.StreamHandler()
        logger_formatter = logging.Formatter(
            fmt='%(levelname)-5s @ %(asctime)s %(name)s: %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
    elif log_mode == 'f':
        logger_handler = logging.FileHandler(log_path)
        logger_formatter = logging.Formatter(
            fmt='%(levelname)-5s @ %(asctime)s %(name)s: %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S'
        )
    else:
        raise KeyError

    logger_handler.setFormatter(logger_formatter)
    logger.addHandler(logger_handler)

    if level not in ('NOTSET', 'INFO'):
        logger.info(f'set log level = {level}')

    return logger


if __name__ == '__main__':
    #  ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG', 'NOTSET'],
    # 一个比一个轻
    # logger = set_logging_level('CRITICAL')  # 只打印 CRITICAL
    # logger = set_logging_level('ERROR')  # DEBUG、INFO、WARNING 没打印
    # logger = set_logging_level('WARNING')  # DEBUG、INFO 没打印
    logger = get_logger('INFO')  # DEBUG 没打印
    # logger = set_logging_level('DEBUG')  # 全打印
    # logger = set_logging_level('NOTSET')  # 全打印
    # CRITICAL > ERROR > WARNING > INFO > DEBUG = NOTSET
    # 也就是说,如果设置为WARNING,就只打印WARNING及左边,的CRITICAL和ERROR
    logger.debug('Python debug')
    logger.info('Python info')
    logger.warning('Python warning')
    logger.error('Python Error')
    logger.critical('Python critical')



from __future__ import absolute_import
import gzip
import sys
from bioat.logger import get_logger

__module_name__ = "bioat.fastxtools"


class FastxTools:
    def __init__(self):
        pass

    @staticmethod
    def _load_fastx_generator(fx, log_level='WARNING'):
        """

        :param str file: path of input <fastq | fastq.gz | fastx | fastx.gz>
        :return: a generator for all reads:
            i.e. print(next(obj)) -> ['header', 'seq', 'info', 'quality']
        :rtype: generator
        """
        logger = get_logger(level=log_level, module_name=__module_name__, func_name=sys._getframe().f_code.co_name)
        f = open(fx, 'rt') if not fx.endswith('.gz') else gzip.open(fx, 'rt')
        # FASTQ @  | FASTA >
        symbol = f.read(1)
        f.close()
        f = open(fx, 'rt') if not fx.endswith('.gz') in fx else gzip.open(fx, 'rt')

        if symbol == '@':
            # FASTQ
            logger.debug('is fastq!')
            read = []
            line = f.readline().rstrip()
            # TODO

            while True:
                if not line:
                    break
                else:
                    if line.startswith('@'):
                        n = len(read)
                        if n == 0:
                            # read == []
                            # 第一次循环开始
                            read.append(line)  # add header!
                            line = f.readline().rstrip()
                        elif n == 4:
                            # read == ['header', 'seq', 'info', 'quality']
                            # ls.append(read)
                            yield read
                            read = []
                            read.append(line)
                            line = f.readline().rstrip()
                        elif n == 3:
                            # TODO
                            """
                            @@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6 # line! 期望它是 header！现在它是 quality
                            [
                                '@Beta12AdemL1C001R00100001768/1',
                                'ATCCCCGTATCTTCACCCCACCACAAACTATTAG',
                                '+',
                            ]'@@IIEIBCE>IC<IBIIIIEAIEIEB<IDECCD6'

                            """
                            read.append(line)
                            line = f.readline().rstrip()

                            if not line.startswith('@'):
                                raise ValueError('The file may be incomplete!')
                        else:
                            f.close()
                            raise ValueError('The file may be incomplete!')

                        # header line!
                        # read.append(line)  # add header !
                    else:
                        # not header line!
                        read.append(line)
                        line = f.readline().rstrip()

            # ls.append(read)
            yield read
            f.close()
            # return ls
        elif symbol == '>':
            # FASTA
            # print('is fastx!')
            ls = []
            read = []
            seq = ''
            # raw_info = [i.rstrip() for i in f.readlines()]
            line = f.readline().rstrip()
            # print(raw_info)

            while True:
                if not line:
                    break
                else:
                    if line.startswith('>'):
                        # 读取 header！
                        n = len(read)

                        if n == 0:
                            # 第一次循环
                            read.append(line)
                            line = f.readline().rstrip()
                        elif n == 1:
                            # 已经有一个 header 了！现在缺 seq
                            read.append(seq)  # add seq line
                            # ls.append(read)
                            yield read
                            read = []  # 重置 read 这个 list
                            read.append(line)
                            line = f.readline().rstrip()
                            seq = ''  # 重置 seq 这个 str
                        else:
                            f.close()
                            raise ValueError('The file may be incomplete!')
                    else:
                        # 读取并添加 seq
                        seq += line
                        line = f.readline().rstrip()

            read.append(seq)
            # ls.append(read)
            yield read
            # return ls
        else:
            f.close()
            raise ValueError('Input line one must starts with @ for FQ or > for FA!')
        f.close()

    def _load_fastx(self):
        if self.iterable:
            return self.__load_fastx_generator()
        else:
            return self.__load_fastx_list()

    def get_headers(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return [i[0][1:] for i in self.fastx]

    def get_sequence(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return [i[1] for i in self.fastx]

    def get_record_dict(self):
        # if self.fastx is None:
        self.fastx = self._load_fastx()

        return {i[0][1:]: i[1] for i in self.fastx}

    # cli subcmd
    def filter_read_contains_N(self, file: str, output=sys.stdout.name):
        """Filter read contains N base in FASTA or FASTQ



        :param file: FASTA | FASTQ file name.
        :param output: FASTA | FASTQ file name file, stdout if not assigned, format as input.
        """
        if self.fastx is None:
            self.fastx = self._load_fastx()

        self.file = file
        output = sys.stdout if output == sys.stdout.name else output

        if isinstance(output, str):
            f = gzip.open(output, 'wt') if output.endswith('.gz') else open(output, 'wt')
        else:
            f = output

        fx = self._load_fastx()

        for read in fx:
            seq = read[1]

            if seq.upper().__contains__('N'):
                continue
            else:
                f.write('\n'.join(read) + '\n')
        f.close()


if __name__ == '__main__':
    # fastx = FastxTools(file='../../data/random_l-180_n-100_hg38_minus-test.fa')
    pass



from __future__ import absolute_import

import dataclasses
import gzip
import os
import sys
from bioat import get_logger
from bioat.lib.libdataclasses import Bed


class BedTools:
    """Bed toolbox."""

    def __init__(self):
        pass

    def random(self, length=200, number=100, fasta=None, seed=0, rm_N=False) -> Bed:
        """Get random region

        :param length: random region length.
        :param number: random region number.
        :param fasta: get random region from which genome, must defined when rm_n=True
        :param seed: random seed
        :param rm_N: remove random region contains N base (It's often the telomere region)
        :return: bed file
        """
        pass



from __future__ import absolute_import
import sys

import pandas as pd
from Bio.Seq import Seq
from Bio.Align import PairwiseAligner
from bioat.lib.libalignment import get_alignment_info
from bioat import get_logger

__module_name__ = 'bioat.lib.libcrispr'

# TARGET_REGIONS_LIB
#     for target_seq alignment
#
# bioat target_seq region_heatmap test_sorted.mpileup.info.tsv test.pdf --get_built_in_target_region
#
# INFO:root:You can use <key> in built-in <target_regions> to represent your target_region:
#         <key>   <target_region>
#         EMX1    GAGTCCGAGCAGAAGAAGAA^NGG^
#         HEK3    GGCCCAGACTGAGCACGTGA^NGG^
#         HEK4    GGCACTGCGGCTGGAGGTGG^NGG^
#         ...

TARGET_SEQ_LIB = {
    "EMX1": "GAGTCCGAGCAGAAGAAGAA^NGG^",
    "HEK3": "GGCCCAGACTGAGCACGTGA^NGG^",
    "HEK4": "GGCACTGCGGCTGGAGGTGG^NGG^",
    "RNF2": "GTCATCTTAGTCATTACCTG^NGG^",
    "VEGFA": "GACCCCCTCCACCCCGCCTC^NGG^",
    "CDKN2A": "^TTTA^GCCCCAATAATCCCCACATGTCA",
    "DYRK1A": "^TTTA^GAAGCACATCAAGGACATTCTAA",
    "ND4": "TGCTAGTAACCACGTTCTCCTGATCAAATATCACTCTCCTACTTACAGGA",
    "ND5.1": "TAGCATTAGCAGGAATACCTTTCCTCACAGGTTTCTACTCCAAAGA",
    "ND6": "TGACCCCCATGCCTCAGGATACTCCTCAATAGCCATCG",
}


def cmp_align_list(aln_a: dict, aln_b: dict):
    """
    INPUT
        like [17, 3, 0, 73.0, 1, 22 'AAGAAGAAGACGAGTCTGCA', '||||||||||||||X|||XX', 'AAGAAGAAGACGAGCCTGAG']
        {
            'match_count': 26,
            'mismatch_count': 3,
            'gap_count': 4,
            'aln_score': 96.0,
            'ref_aln_start': 0,
            'ref_aln_end': 32,
            'alignment': {
                'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT',
                'aln_info':      '--..|.|||||||||||||||||||||||--||',
                'target_seq':    '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'
            }
        }
    HELP
        compare function for align_list
    """

    # sort reason score -> gap -> mismatch -> match
    def sign_value(x):
        """
        HELP
            sign function
        """
        if x > 0:
            return 1
        elif x < 0:
            return -1
        else:
            return 0

    align_alphabet_score = {"-": 0, ".": 1, "|": 2}
    sort_keys = ['alignment_score', 'gap_count', 'mismatch_count', 'match_count']
    sort_rev_state_list = [True, False, False, True]

    for idx, key in enumerate(sort_keys):
        if aln_a[key] - aln_b[key] == 0:
            continue
        else:
            if sort_rev_state_list[idx]:
                return -1 * sign_value(aln_a[key] - aln_b[key])
            else:
                return sign_value(aln_a[key] - aln_b[key])

    aln_info_a = aln_a['alignment']['aln_info']
    aln_info_b = aln_b['alignment']['aln_info']

    for idx, char_a in enumerate(aln_info_a):
        if idx <= (len(aln_info_b) - 1):
            value_a = align_alphabet_score[char_a]
            value_b = align_alphabet_score[aln_info_a[idx]]
            return sign_value(value_a - value_b)
        else:
            pass


def run_target_seq_align(
        ref_seq: Seq,
        target_seq: Seq,
        aligner: PairwiseAligner,
        PAM: dict = None,
        log_level='WARNING'
) -> list:
    """global alignment for target_seq

    :param ref_seq: a Seq object from BioPython, reference sequence for the targeted deep sequencing
    :param target_seq: a Seq object from BioPython, target region sequence for the editing window, usually the sgRNA
        sequencing without PAM
    :param aligner: an obj from BioPython pairwise alignment
    :param PAM: a dict like {'PAM': 'AGG', 'position': 20, 'weight': 1.0},
        position is the insertion site of target_seq,
        weight is the priority, PAM alignment score will multiply by this weight.
    :return: a list: list contains some dict, each one is an alignment result.
    [
        {
            'match_count': 26,
            'mismatch_count': 3,
            'gap_count': 4,
            'aln_score': 96.0,
            'ref_aln_start': 0,
            'ref_aln_end': 32,
            'alignment': {
                'reference_seq': 'GGCACTGCGGCTGGAAAAAAAAAAAAAAA--GT',
                'aln_info':      '--..|.|||||||||||||||||||||||--||',
                'target_seq':    '--GGCAGCGGCTGGAAAAAAAAAAAAAAAAGGT'
            }
        },
        ...
    ]

    """
    # set logger
    logger = get_logger(level=log_level, module_name=__module_name__, func_name=sys._getframe().f_code.co_name)

    if PAM:
        # considering PAM
        # DNA info https://www.bioinformatics.org/sms/iupac.html
        alignments_pam = aligner.align(ref_seq, Seq(PAM['PAM']))
        # use pam to find aim sequence on ref and then use target_seq to align them and give it a higher priority
        dt_PAMs = {}

        for idx, alignment in enumerate(alignments_pam):
            alignment_analysis_result = get_alignment_info(alignment, reverse=False)
            logger.debug(f'alignment_analysis_result for PAM:\n{alignment_analysis_result}')
            ref_aln_start = alignment_analysis_result['ref_aln_start']
            ref_aln_end = alignment_analysis_result['ref_aln_end']
            aln_score = alignment_analysis_result['aln_score'] * PAM['weight']
            dt_PAMs[f'{ref_aln_start}_{ref_aln_end}'] = dict(aln_score=aln_score * PAM['weight'])

        logger.debug(f'find PAMs on reference: {dt_PAMs}')

        # add PAM left or right or any position
        target_seq = target_seq[:PAM['position']] + Seq(PAM['PAM']) + target_seq[PAM['position']:]
        logger.info(f'find PAM on target_seq, target_seq is updated to: {str(target_seq)}')

    # alignment part
    alignments = aligner.align(ref_seq, target_seq)
    # parse alignment
    alignment_results = []

    for alignment in alignments:
        alignment_analysis_result = get_alignment_info(alignment, reverse=False)
        logger.debug(f'alignment_analysis_result:\n{alignment_analysis_result}')
        alignment_results.append(alignment_analysis_result)

    if len(alignment_results) == 1:
        logger.debug(f'find 1 alignment result:\n{alignment_results}')
        return alignment_results[0]
    else:
        logger.debug(f'find {len(alignment_results)} alignment results:\n{alignment_results}')
        df = pd.DataFrame(alignment_results)
        # if contain multiple alignment result with the same score, keep the best one;
        # sort reason score -> gap -> mismatch -> match
        df.sort_values(
            by=['aln_score', 'gap_count', 'mismatch_count', 'match_count'],
            ascending=[False, True, True, False],
            inplace=True
        )
        return df.iloc[0, :].to_dict()
